-- How the IO monad works

module MyIO(MyIO, myPutChar, myGetChar, convert) where
import Control.Applicative
import Control.Monad

type Input = String
type Remainder = String
type Output = String

data MyIO a  =  MyIO (Input -> (a, Remainder, Output))

apply :: MyIO a -> Input -> (a, Remainder, Output)
apply (MyIO f) inp  =  f inp

myPutChar :: Char -> MyIO ()
myPutChar ch  =  MyIO (\inp -> ((), inp, [ch]))

myGetChar :: MyIO Char
myGetChar =  MyIO (\(ch:rem) -> (ch, rem, []))

-- *******************************************************
-- DIFFERENCE WITH RESPECT TO LECTURES
-- The following instance declarations for Functor and Applicative, and
-- a change to the instance declaration for Monad, are required because
-- of a change to the way that Monad is defined in Haskell 7.10+.
-- see https://wiki.haskell.org/Functor-Applicative-Monad_Proposal
-- and https://ghc.haskell.org/trac/ghc/wiki/Migration/7.10
-- *******************************************************

instance Functor MyIO where
  fmap = liftM
 
instance Applicative MyIO where
  pure x  = MyIO (\inp -> (x, inp, ""))
  (<*>)   = ap

instance Monad MyIO where
  m >>= k   =  MyIO (\inp ->
                 let (x, rem1, out1) = apply m inp in
                 let (y, rem2, out2) = apply (k x) rem1 in
                 (y, rem2, out1++out2))

convert :: MyIO () -> IO ()
convert m  =  interact (\inp ->
                let (x, rem, out) = apply m inp in
                out)




